/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

////////////////////////////////////////////////////////////////////////////
//
// AtomicPart Methods
//
////////////////////////////////////////////////////////////////////////////

#include <libc.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <stdio.h>
#include <strings.h>
#include <List.h>
#include "oo7.h"
#include "globals.h"
#include "VarParams.h"
#include "GenParams.h"


extern int	debugMode;

extern char*	types[NumTypes];

extern long getListIndex (List* aList, Entity* anEntity);

////////////////////////////////////////////////////////////////////////////
//
// AtomicPart Constructor
//
////////////////////////////////////////////////////////////////////////////

AtomicPart::AtomicPart(int ptId, CompositePart *cp)
{
    if (debugMode) {
        printf("AtomicPart::AtomicPart(ptId = %d)\n", ptId);
    }
    initDirectType(AtomicPartType);     

    // initialize internal state of new part
    id = ptId;
    int typeNo = (int) (random() % NumTypes);
    strncpy(type, types[typeNo], TypeSize);
    buildDate = MinAtomicDate +
        (int) (random() % (MaxAtomicDate-MinAtomicDate+1));
    x = (int) (random() % XYRange);
    y = (int) (random() % XYRange);
    // fill in a random document id (for query 9)
    docId = (int) random() % TotalCompParts + 1;

    // initialize space for connections to other parts
    fromListReference.Init( new (oo7Group,OC_list) List(CONType));

    toListReference.Init( new (oo7Group,OC_list) List(CONType));

    // initialize reference to composite part.
    compPartReference.Init(cp);

    // add to hash table on id.
    AllPartsById->Insert(id, this);

    // add to B+ tree on date.
    AllPartsByDate->Insert(buildDate, this);
}

// Additional constructor required by Ontos.
AtomicPart::AtomicPart(APL *theAPL) : Object(theAPL) {}

// destructor
AtomicPart::~AtomicPart()
{
    Destroy (FALSE);
}

// Destroy
void AtomicPart::Destroy (OC_Boolean aborted)
{
    Connection* 	cn;
    ListIterator*	connI;
    AtomicPart* 	ap;
    long 		pos;

    // deallocate the "to" list and the connections 
    to()->deallocateCluster(TRUE);

    // deallocate the "from" list if active
    if (fromListReference.isActive(this)) {
	List* fromList = from();
	delete fromList;
    }

    if (aborted)
	Object::Destroy (aborted);
}

// deleteObject
void AtomicPart::deleteObject(OC_Boolean deallocate)
{
    Connection* 	cn;
    ListIterator*	connI;
    AtomicPart* 	ap;
    long 		pos;

    // remove part from extent of all atomic parts
    AllPartsById->Remove(id, this);

    // put the set
// GKK
#ifndef MARKMODIFIED
    AllPartsById->putObject();
#else
    AllPartsById->markModified();
#endif

    // remove part from extent of all atomic parts
    AllPartsByDate->Remove (buildDate, this);

    // put the set
// GKK
#ifndef MARKMODIFIED
    AllPartsByDate->putObject();
#else
    AllPartsByDate->markModified();
#endif

    // establish iterator over connected parts in "to" direction
    connI = new ListIterator(to());
    while (connI->moreData())
    {
         cn = (Connection *)(Entity *)(*connI)();
         ap = cn->to();  // get the associated atomic part

         pos = getListIndex(ap->from(), cn);
         if (pos < 0) { // not found
             printf("in AtomicPart::deleteObject, couldn't find connection %x.\n",
		     (int) cn);
             OC_transactionAbort();
             exit(1);
         } 
         // know cn is in list
         ap->from()->Remove(pos);

         // make sure change makes it to DB
         ap->from()->putObject();

         // delete and deallocate the connection object.
	 cn->deleteObject(deallocate); // delete the connection object
    }
    delete connI;

    to()->deleteObject(FALSE); // delete the "to" list

    // establish iterator over connected parts in "from" direction
    connI = new ListIterator(from());
    while (connI->moreData())
    {
         cn = (Connection *)(Entity *)(*connI)();
         ap = cn->from();  // get the associated atomic part
         pos = getListIndex(ap->to(), cn);
         if (pos < 0) { // not found
             printf("in AtomicPart destructor, couldn't find connection %x.\n",
                     (int) cn);
             OC_transactionAbort();
             exit(1);
         } 
         // know cn is in list
	 ap->to()->Remove(pos);
 
         // make sure changes make it to DB
         ap->to()->putObject();
 
         // delete and deallocate the connection object.
	 cn->deleteObject(deallocate); // delete the connection object
    }
    delete connI;

    from()->deleteObject(FALSE); // delete the "from" list

    Object::deleteObject (deallocate);
}


////////////////////////////////////////////////////////////////////////////
//
// getDirectType required by ONTOS.
//
////////////////////////////////////////////////////////////////////////////

Type* AtomicPart::getDirectType() {
//    return (Type*)OC_lookup("AtomicPart");
    return APType;
}

////////////////////////////////////////////////////////////////////////////
//
// AtomicPart swapXY Method for use in update traversals
//
////////////////////////////////////////////////////////////////////////////

void AtomicPart::swapXY()
{
    if (debugMode) {
        printf("                    AtomicPart::swapXY(ptId = %d, x = %d, y = %d)\n", id,x,y);
    }

    // exchange X and Y values

    int tmp = x;
    x = y;
    y = tmp;

    // put the change into the DB
    this->putObject();

    if (debugMode) {
        printf("                    [did swap, so x = %d, y = %d]\n", x, y);
    }
}


////////////////////////////////////////////////////////////////////////////
//
// AtomicPart toggleDate Method for use in update traversals
//
////////////////////////////////////////////////////////////////////////////

void AtomicPart::toggleDate()
{
    if (debugMode) {
        printf("                    AtomicPart::toggleDate(ptId = %d, buildDate = %d)\n",
	       id, buildDate);
    }

    // delete from index first
    AllPartsByDate->Remove(buildDate,this);

    // increment build date if odd, decrement it if even

    if (buildDate % 2) {
	// odd case
	buildDate++;
    } else {
	// even case
	buildDate--;
    }
    // update the index
    AllPartsByDate->Insert(buildDate,this);

    // put changes to the object to the DB
    this->putObject();

    if (debugMode) {
        printf("                    [did toggle, so buildDate = %d]\n", buildDate);
    }
}


////////////////////////////////////////////////////////////////////////////
//
// AtomicPart DoNothing Method 
//
////////////////////////////////////////////////////////////////////////////

int RealWork; // set to one to make DoNothing do work.
int WorkAmount; // amount of work to do

void AtomicPart::DoNothing()
{
    struct timeval tv;
    struct timezone tz;
    int ignore;

    int i;
    if (id < 0) 
        printf("DoNothing - negative id\n");


    if (debugMode) {
        printf("==> DoNothing(x = %d, y = %d, type = %s)\n", x, y, type);
    }

    if (RealWork) {
            for (i = 0; i < WorkAmount; i++) {
               gettimeofday(&tv, &tz);
            }
    }
}


/*  $Version: AtomicPart.C 1.5 93/02/25 10:08:37 Exp $ */

