/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


////////////////////////////////////////////////////////////////////////////
//
// CompositePart Methods
//
////////////////////////////////////////////////////////////////////////////

#include <libc.h>
#include <stdio.h>
#include <strings.h>
#include <Dictionary.h>
#include <List.h>
#include <Set.h>
#include <Entity.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "VarParams.h"
#include "baidlist.h"

extern BAIdList* private_cp;
extern BAIdList* shared_cp;

extern int	nextAtomicId;
extern int	debugMode;
extern char*	types[NumTypes];


extern long getListIndex (List* aList, Entity* anEntity);

////////////////////////////////////////////////////////////////////////////
//
// CompositePart Constructor
//
////////////////////////////////////////////////////////////////////////////

CompositePart::CompositePart(int cpId)
{
    int 	i, from, to;
    int 	atomicid;
    AtomicPart** atomicParts;
    Connection*	cn;
    BaseAssembly* ba;
    char	title[TitleSize];

    if (debugMode) {
        printf("CompositePart::CompositePart(cpId = %d)\n", cpId);
    }

    atomicParts = new AtomicPart*[NumAtomicPerComp];
    initDirectType(CompositePartType);     

    // add comp part to extent of all comp parts
    AllCompPartsById->Insert(cpId, this);
// GKK
#ifndef MARKMODIFIED
    AllCompPartsById->putObject();
#else
    AllCompPartsById->markModified();
#endif

    // initialize the simple stuff
    id = cpId;
    int typeNo = (int) (random() % NumTypes);
    strncpy(type, types[typeNo], TypeSize);

    // for the build date, decide if this part is young or old, and then
    // randomly choose a date in the required range

    if (cpId % YoungCompFrac == 0) {
        // young one
        if (debugMode) {
            fprintf(stderr, "(young composite part, id = %d.)\n", id);
        }
        buildDate = MinYoungCompDate +
                    (int)(random() % (MaxYoungCompDate - MinYoungCompDate + 1));
    } else {
        // old one
        if (debugMode) {
            fprintf(stderr, "(old composite part, id = %d.)\n", id);
        }
        buildDate = MinOldCompDate +
                    (int)(random() % (MaxOldCompDate - MinOldCompDate + 1));
    }

    // initialize reference to private list
    sharReference.Reset(
        new (oo7Group,OC_list) List(BAType));

    // initialize reference to shared association
    privReference.Reset(
        new (oo7Group,OC_list) List(BAType));
    
    // initialize reference to parts association
    partsReference.Reset(
        new (oo7Group,OC_set) Set(APType));

    // initialize the documentation (indexed by its title) ...
    sprintf(title, "Composite Part %08d", cpId);

    // initialize reference to the document.
    documentReference.Reset(
	new (oo7Group, DOCType) Document(cpId, this),this);

    // put the document in the database.
    // and deallocate.
    documentation()->putObject();

    // insert title into document index
    AllDocumentsByTitle->Insert(title, documentation());

    // insert id into document index
    AllDocumentsById->Insert(cpId, documentation());

    // now create the atomic parts (indexed by their ids) ...
    for (i = 0; i < NumAtomicPerComp; i++) 
    {
	atomicid = nextAtomicId + i;

	// create a new atomic part
	// the AtomicPart constructor takes care of setting up
	// the back pointer to the containing CompositePart()
	atomicParts[i] = new (oo7Group, APType) AtomicPart(atomicid,this);

	// connect the atomic part to the composite part
	parts()->Insert(atomicParts[i]);

	// first atomic part is the root part
	if (i == 0) rootPartReference.Reset(atomicParts[i]);
    }

    // put the parts association.
    parts()->putObject();
    // put the usedInPriv list (even though it is empty)
    usedInPriv()->putObject();
    // put the usedInShar list (even though it is empty)
    usedInShar()->putObject();

    // ... and then wire them semi-randomly together (as a ring plus random
    // additional connections to ensure full part reachability for traversals)

    for (from = 0; from < NumAtomicPerComp; from++) {
	for (i = 0; i < NumConnPerAtomic; i++) {
	    if (i == 0) {
	        to = (from + 1) % NumAtomicPerComp;
	    } else {
	        to = (int) (random() % NumAtomicPerComp);
	    }
	    cn = new (oo7Group, CONType) 
			Connection(atomicParts[from], atomicParts[to]);

if (cn == NULL)  {
printf("cn was null!\n");
fflush(stdout);
}

            // put the new connection to the database.
            // and deallocate the object.
            cn->putObject();
	}
    }

    // put all new atomic Parts into the DB.  (ONTOS specific.)
    // Deallocate them as you go...
    for (i = 0; i < NumAtomicPerComp; i++) {
        atomicParts[i]->putObject();
    }
    nextAtomicId += NumAtomicPerComp;

    // finally insert this composite part as a child of the base
    // assemblies that use it

    // first the assemblies using the comp part as a shared component
    shared_cp[cpId].initScan();

    if (debugMode) {
	    printf("Starting a scan on shared_cp[%d].\n", cpId);
    }

    // get the first base assembly
    ba = shared_cp[cpId].next();
    while (ba != NULL)
    {
         if (debugMode) {
             printf("\tAdding ba %d.\n", ba->id);
        }
        // add this composite part cp the list of shared parts used
        // in the assembly
        ba->componentsShar()->Insert(this);
	ba->componentsShar()->putObject();
	ba->putObject();
        usedInShar()->Insert(ba);

        // continue with next base assembly
        ba = shared_cp[cpId].next();
    }

    // next the assemblies using the comp part as a private component
    private_cp[cpId].initScan();

    // get the first base assembly
    ba = private_cp[cpId].next();
    while (ba != NULL)
    {
        // add this composite part to the list of private parts used
        // in the assembly
        ba->componentsPriv()->Insert(this);
	ba->componentsPriv()->putObject();
	ba->putObject();
        usedInPriv()->Insert(ba);
        // continue with next base assembly
        ba = private_cp[cpId].next();
    }
}

// Additional constructor required by Ontos.
CompositePart::CompositePart(APL *theAPL) : Object(theAPL) {}

////////////////////////////////////////////////////////////////////////////
//
// getDirectType required by ONTOS.
//
////////////////////////////////////////////////////////////////////////////

Type* CompositePart::getDirectType() {
    return CPType;
}




// destructor
CompositePart::~CompositePart()
{
    Destroy (FALSE);
}

inline deallocateObject (Object* obj) { delete obj; }

// Destroy
void CompositePart::Destroy(OC_Boolean aborted)
{
    // deallocate document
    if (documentReference.isActive(this))
	deallocateObject (documentation());

    // deallocate list of private components
    if (privReference.isActive(this))
	deallocateObject (usedInPriv());

    // deallocate list of shared components
    if (sharReference.isActive(this))
	deallocateObject (usedInShar());
	
    // deallocate all atomic parts that compose this composite part
    if (partsReference.isActive(this))
	parts()->deallocateCluster(TRUE);

    if (aborted)
	Object::Destroy (aborted);
}

// deleteObject
void CompositePart::deleteObject(OC_Boolean deallocate)
{
    ListIterator* baseI;  // iterator for traversing subassemblies
    BaseAssembly* baH;
    SetIterator*  atomI;  // iterator for traversing atomic parts
    AtomicPart*	  apH;
    AtomicPart*	  apH2;
    Document* 	  docH;
    char	  docTitle[40];
    long          pos;
    List*	 listx;
    int		  docId;

    strcpy(docTitle,documentation()->title);
    docId = documentation()->id;

    docH = (Document*) (Entity*) AllDocumentsByTitle->Remove(docTitle);
    docH = (Document*) (Entity*) AllDocumentsById->Remove(docId);
    documentation()->deleteObject(deallocate);

    // walk through usedInPriv list deleting all backward references
    // to this composite part from base assemblies of which it is a part
    baseI = new ListIterator(usedInPriv());  // establish baseassembly iterator
    while (baseI->moreData())
    {
        baH = (BaseAssembly *)(Entity *)(*baseI)(); 
        pos = getListIndex(baH->componentsPriv(), this);
        if (pos < 0) {
            printf("In CompositePart destructor, bad delete.\n");
            OC_transactionAbort();
            exit(1);
        }
	baH->componentsPriv()->Remove(pos);
        // make sure change gets to DB.
        baH->componentsShar()->putObject();
    }
    listx = usedInPriv();
    if (listx != NULL) listx->deleteObject(deallocate); // delete and deallocate
                                      // the association object
    delete baseI;

    // walk through usedInShar list deleting all backward references
    // to this composite part from base assemblies of which it is a part
    baseI = new ListIterator(usedInShar());  // establish baseassembly iterator
    while (baseI->moreData())
    {
        baH = (BaseAssembly *)(Entity *)(*baseI)(); 
        pos = getListIndex(baH->componentsShar(), this);
        if (pos < 0) {
            printf("In CompositePart destructor, bad delete.\n");
            OC_transactionAbort();
            exit(1);
        }
	baH->componentsShar()->Remove(pos);
        // make sure change gets to DB.
        baH->componentsShar()->putObject();
    }
    listx = usedInShar();
    if (listx != NULL) listx->deleteObject(deallocate); // delete and deallocate
		                      // the association object
    delete baseI;

    // delete all atomic parts that compose this composite part
    atomI = new SetIterator(parts());  // establish atomic part iterator
    while (atomI->moreData())
    {
        apH = (AtomicPart*)(Entity *)(*atomI)();
        // remove from id index.
	apH2 = (AtomicPart*)(Entity*) AllPartsById->Remove(apH->id);
/*
	if (apH != apH2) printf("index error when deleting atomic parts \n");
*/

        // remove from date index.
        AllPartsByDate->Remove(apH->id, apH);

	apH->deleteObject(deallocate); // the atomic part destructor cleans up
		                  // connection objects
    }
    delete atomI;

    // write the index changes to disk.
// GKK
#ifndef MARKMODIFIED
    AllPartsById->putObject();
    AllPartsByDate->putObject();
#else
    AllPartsById->markModified();
    AllPartsByDate->markModified();
#endif

    parts()->deleteObject(deallocate);  // delete parts association


    // finally remove from the extent
    AllCompPartsById->Remove(this->id, this);  

// GKK
#ifndef MARKMODIFIED
    AllCompPartsById->putObject();   // write changes to DB.
#else
    AllCompPartsById->markModified();
#endif

    Object::deleteObject (deallocate);
}

/*  $Version: CompositePart.C 1.4 93/02/25 10:08:39 Exp $ */

