/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/

#include <libc.h>
#include <stdio.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <strings.h>
#include <Transaction.h>
#include <stdlib.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "VarParams.h"
#include "stats.h"
#include "baidlist.h"

extern "C" char* getenv(const char*);

// This is a gendb hacked to generate the large db in pieces.
// It takes two parameters.  The starting composite partid and
// the ending composite part id.  Modules and assemblies are generated
// only when the second parameter equals the total number of
// composite parts

//////////////////////////////////////////////////////////////////
//
// Global Variables, etc., for Benchmark Database Generation
//
//////////////////////////////////////////////////////////////////

struct rusage startUsage, endUsage;
struct timeval startWallTime;
struct timeval endWallTime;
struct timezone ignoreTimeZone;

extern SetParams(char* configFileName);
extern void InitGlobals();
extern void EstablishGlobals();
extern void createDummy(int cnt);

extern double ComputeWallClockTime(struct timeval *startWallTime, 
	           struct timeval *endWallTime);

extern double ComputeUserTime(struct rusage *startUsage, 
	           struct rusage *endUsage);

extern double ComputeSystemTime(struct rusage *startUsage, 
	           struct rusage *endUsage);

int	nextAtomicId=1;
int	nextCompositeId=1;
int	nextComplexAssemblyId=1;
int	nextBaseAssemblyId=1;
int	nextModuleId=1;
int	debugMode = FALSE;

char    *types[NumTypes] = {
		"type000", "type001", "type002", "type003", "type004",
		"type005", "type006", "type007", "type008", "type009"
	};

// data structures to keep track of which composite parts each base
// assemblies uses.  This is done so that we can generate the base
// assemblies first and then the composite parts.  Why bother?
// In particular, if you did the composite parts first and the base
// assemblies, there would be no need for these data structures.
// This second approach worked fine for the small and medium databases
// but not for the large as the program seeked all over the disk.

BAIdList* private_cp;
BAIdList* shared_cp;

//////////////////////////////////////////////////////////////////
//
// Do Benchmark Data Generation
//
//////////////////////////////////////////////////////////////////

void gendb(int startCompId, int endCompId)
{
        int 		id;
	char 		moduleName[40];
	Module*		moduleH;
	CompositePart*	cp;

        // First generate the desired number of modules
        if (startCompId == 1) 
	{ 
		// first run, generate modules
		while (nextModuleId <= TotalModules) 
		{
		    int id = nextModuleId++;
		    printf("about to generate module %d\n",id);
		    fflush(stdout);
		    moduleH = new (oo7Group, MODType) Module (id);
	            // add module to index.
	            sprintf(moduleName, "Module %08d", id);
		    AllModulesByName->Insert(moduleName, moduleH); 
		    moduleH->putObject();
		}
	}

	printf("Made modules, now on to composite parts.\n");
	fflush(stdout);

        // now create a batch of composite parts.
	nextCompositeId = startCompId;
	nextAtomicId = (startCompId-1)*NumAtomicPerComp + 1;
	while (nextCompositeId <= endCompId) 
	{
	    id = nextCompositeId;
	    cp = new (oo7Group, CPType) 
			CompositePart(id);
            // put the object to the database 
	    cp->putObject();
	    nextCompositeId++;
	    if ((nextCompositeId % 100) == 0) {
		printf("Made composite part %d.\n", nextCompositeId);
	        fflush(stdout); 
            }
        }
	printf("made composite parts %d to %d.\n", startCompId, endCompId);
        fflush(stdout);

	// finally flush out all the root dictionary objects
	AllBaseAssemblies->putObject();  
        AllModulesByName->putObject();
	AllCompPartsById->putObject();
	AllDocumentsByTitle->putObject();
	AllDocumentsById->putObject();
	AllPartsById->putObject();
	AllPartsByDate->putObject();
	oo7Group->putObject();

	// Commit the transaction
        OC_transactionCommit();

        // compute and report wall clock time
        gettimeofday(&endWallTime, &ignoreTimeZone);       
	getrusage(RUSAGE_SELF, &endUsage);

	// Print out some useful information about what happened
	printf("=== DONE CREATING DATABASE, TOTALS WERE ===\n");
	printf("    # atomic parts\t\t%d\n", nextAtomicId-1);
	printf("    # composite parts\t\t%d\n", nextCompositeId-1);
	printf("    # complex assemblies\t%d\n", nextComplexAssemblyId-1);
	printf("    # base assemblies\t\t%d\n", nextBaseAssemblyId-1);
	printf("    # modules\t\t\t%d\n", nextModuleId-1);

	//////////////////////////////////////////////////////////////////
	//
	// Shutdown DB
	//
	//////////////////////////////////////////////////////////////////

	printf("Wall-Clock time to generate database: %f seconds.\n", 
	             ComputeWallClockTime(&startWallTime, &endWallTime));
	fprintf(stdout, "(CPU: %f seconds.)\n", 
	                ComputeUserTime(&startUsage, &endUsage) +
			ComputeSystemTime(&startUsage, &endUsage));
}


main(int argc, char** argv)
{
	if (argc < 4) 
	{
	      fprintf(stderr, 
                  "Usage: %s <configFileName> <startCompId> <endCompId> [-d]\n", argv[0]);
	      exit(1);
	}
	// See if debug mode is desired
	if ((argc == 5) && (strcmp(argv[4], "-d") == 0)) debugMode = TRUE;
	else  debugMode = FALSE;
	SetParams(argv[1]);

	// allocate space for BAlists
	shared_cp = new BAIdList[TotalCompParts+1];
	private_cp = new BAIdList[TotalCompParts+1];

	// get starting composite partID.
	int startCompID  = atoi(argv[2]);
	int endCompID  = atoi(argv[3]);
	printf("generating composite parts %d to %d\n", startCompID, endCompID);
        fflush(stdout);

        // get wall clock time
        gettimeofday(&startWallTime, &ignoreTimeZone);

	// get starting usage values.
	getrusage(RUSAGE_SELF, &startUsage);

        // GKK 
        OC_groupProtectionDisable();

        // open the OO7 database
        OC_open("oo7db");

        // start a transaction
        OC_transactionStart();

        APType = (Type *)OC_lookup("AtomicPart");
        CPType = (Type *)OC_lookup("CompositePart");
        ASSType = (Type *)OC_lookup("Assembly");
        CAType = (Type *)OC_lookup("ComplexAssembly");
        BAType = (Type *)OC_lookup("BaseAssembly");
        CONType = (Type *)OC_lookup("Connection");
        MANType = (Type *)OC_lookup("Manual");
        DOCType = (Type *)OC_lookup("Document");
        MODType = (Type *)OC_lookup("Module");

	// create all the indexes and extents needed
        if (startCompID == 1) InitGlobals();
	else 
	{
		EstablishGlobals();
		printf("restoring baidlists\n");
		shared_cp->restore(TotalCompParts+1, "shared_cp.data");
		private_cp->restore(TotalCompParts+1, "private_cp.data");
		printf("baidlists restore completed\n");
	}

        // generate the database
	gendb(startCompID, endCompID); 

        if (startCompID == 1) 
	{
		printf("\nBeginning save of baidlists\n");
		shared_cp->save(TotalCompParts+1, "shared_cp.data");
		private_cp->save(TotalCompParts+1, "private_cp.data");
		printf("\nCompleted save of baidlists\n");
	}

        // close the database
        OC_close();
}

/*  $Version: GenDB.C 1.4 93/02/25 10:08:41 Exp $ */

