/********************************************************/
/*                                                      */
/*               OO7 Benchmark                          */
/*                                                      */
/*              COPYRIGHT (C) 1993                      */
/*                                                      */
/*                Michael J. Carey 		        */
/*                David J. DeWitt 		        */
/*                Jeffrey Naughton 		        */
/*               Madison, WI U.S.A.                     */
/*                                                      */
/*	         ALL RIGHTS RESERVED                    */
/*                                                      */
/********************************************************/


#include <libc.h>
#include <string.h>
#include <stdio.h>
#include <List.h>

#include "oo7.h"
#include "globals.h"
#include "GenParams.h"
#include "BenchParams.h"
#include "VarParams.h"
#include "PartIdList.h"



//////////////////////////////////////////////////////////////////
//
// Reorganization Test #1 - Visit all composite parts, deleting
// and then newly inserting 1/2 of each composite part's atomic
// parts graph.
//
//////////////////////////////////////////////////////////////////


extern char*	types[NumTypes];
extern int	debugMode;


//////////////////////////////////////////////////////////////////
//
// Routine to do Reorg1
//
//////////////////////////////////////////////////////////////////


int reorg1()
{

    CompositePart* cp;

    // iterate through all the composite parts, deleting/inserting
    // half of an atomic parts graph worth of parts and counting up
    // the number of atomic parts processed in this way

    int count = 0;

    // create an Ontos "DictionaryIterator"
    DictionaryIterator *compPartItr;

    // creating it with defaults means the iterator will
    // range over all composite parts in the iterator.
    compPartItr = new DictionaryIterator(AllCompPartsById);


    while (compPartItr->moreData())
    {
        // crystal-clear Ontos syntax follows
        cp = (CompositePart*)(Entity*)(*compPartItr)();

        if (debugMode) {
            printf("    In Reorg1, CompPartId = %d:\n", cp->id);
        }
        count += cp->reorg1();
    }
    delete compPartItr;
    return count;
}



//////////////////////////////////////////////////////////////////
//
// Composite Part Methods for Reorg1
//
//////////////////////////////////////////////////////////////////


int CompositePart::reorg1()
{
    AtomicPart* 	atomH;
    Connection* 	cn;
    ListIterator*	connI;
    int partId, i;

    if (debugMode) {
        printf("CompositePart::Reorg1(id = %d)\n", id);
    }

    // start by getting pointers to all of the atomic parts, in increasing
    // part id order, in preparation for removing/recreating half of them

    AtomicPart** atom = new AtomicPart* [NumAtomicPerComp];
    int lowId  = rootPart()->id;
    int highId = lowId + NumAtomicPerComp - 1;

    partId = lowId;
    for (i = 0; i < NumAtomicPerComp; i++) 
    {

	atomH = (AtomicPart*) (Entity*) AllPartsById->Remove(partId);
	if (atomH == NULL) {
            fprintf(stderr, "ERROR: Unable to find atomic part %d\n", partId);
            OC_transactionAbort();
            exit(1);
        }
	atom[i] = atomH;
	partId++;
    }


    // now do the deletions of half of the atomic parts, keeping the first
    // half and targeting the second half for simplicity (to avoid the root
    // part and to enable a simple computation to discriminate between parts
    // to be deleted and those to be retained);
    // before doing the deletions, go through and remember all the connections
    // that will be lost when the targeted parts are deleted;
    // after all that, recreate the parts and reconnect them to everything!

    // first compute the range of impact

    int amtToChange, firstToChange, midId;
    amtToChange = NumAtomicPerComp / 2;
    firstToChange = NumAtomicPerComp - amtToChange;
    midId = highId - amtToChange + 1;

    // next record the relevant connections.

    // allocate space to hold the connected parts
    PartIdList* toParts = new PartIdList [NumAtomicPerComp];

    for (i = 0; i < NumAtomicPerComp; i++) 
    {
	// for each of the atomic parts in the atom[] array
	// scan through the parts it is connected to
        connI = new ListIterator(atom[i]->to());
        while (connI->moreData()) 
	{
            cn = (Connection*)(Entity*)(*connI)();
            atomH = cn->to();  // get the associated atomic part
	    // if either the part or the connected part have id's
	    // greater than midId, save the id of part in the toParts 
	    // array
	    if ((atom[i]->id >= midId) || (atomH->id >= midId)) {
		toParts[i].insert(atomH->id);
	    }
        }
	delete connI;
    }

    // now do the actual part deletions
    int count = 0;
    for (i = firstToChange; i < NumAtomicPerComp; i++) 
    {
	if (debugMode) {
            printf("\tReorg1: Deleting atomic part %d\n", atom[i]->id);
	}
	if (atom[i] != NULL) atom[i]->deleteObject();  
	else printf("was going to call delete on NULL atom[%d]\n",i);
	count++;
    }
    if (debugMode) {
            printf("\tReorg1: Deleted all atomic parts\n");
    }
    

//    Type* atomicPartType = (Type*) OC_lookup ("AtomicPart");
    Type* atomicPartType = APType;

    // time to (approximately) recreate the deleted parts

    partId = midId;
    for (i = firstToChange; i < NumAtomicPerComp; i++) 
    {
	atom[i] = new (oo7Group, atomicPartType) AtomicPart (partId, this);
	if (debugMode) {
            printf("\tReorg1: Recreated atomic part %d\n", atom[i]->id);
	}
	partId++;
    }

//    Type* connectionType = (Type*) OC_lookup ("Connection");
    Type* connectionType = CONType;

    // finally, connect all of the parts back up again

    for (i = 0; i < NumAtomicPerComp; i++) 
    {
	while (!toParts[i].empty()) {
	    int to = toParts[i].remove() - lowId;
	    cn = new (oo7Group, connectionType) Connection (atom[i], atom[to]);
            // put this new connection into the db.
            cn->putObject();
	}
    }

    // put the new atomic parts into the database.
    for (i = 0; i < NumAtomicPerComp; i++) 
    {
        atom[i]->putObject();
    }


    delete atom;
    delete toParts;

    return count;
}


/*  $Version: Reorg1.C 1.3 93/02/25 10:08:47 Exp $ */

